'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var FS = _interopDefault(require('fs'));
var Path = _interopDefault(require('path'));
var vueVirtualTextdocument = require('@vuedx/vue-virtual-textdocument');
var vscodeUri = require('vscode-uri');

const PATCHED_METHODS = Symbol('Vue Patched Methods');
function tryPatchMethod(target, methodName, createOverride) {
    const patched = target[PATCHED_METHODS] || [];
    if (patched.includes(methodName))
        return;
    let fn = target[methodName];
    if (typeof fn === 'function') {
        try {
            fn = fn.bind(target);
        }
        catch (_a) {
            // - ignore
        }
    }
    target[methodName] = createOverride(fn);
    patched.push(methodName);
    target[PATCHED_METHODS] = patched;
}

function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
class ProxyDocumentStore extends vueVirtualTextdocument.DocumentStore {
    get(fileNameOrUri) {
        return super.get(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    has(fileNameOrUri) {
        return super.has(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    set(fileNameOrUri, document) {
        return super.set(vueVirtualTextdocument.asUri(fileNameOrUri), document);
    }
    delete(fileNameOrUri) {
        return super.delete(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
}
class PluginContext {
    constructor(typescript) {
        this.typescript = typescript;
        this.watchers = new Map();
        this.store = new ProxyDocumentStore((uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            const content = this.typescript.sys.readFile(fileName) || '';
            return vueVirtualTextdocument.VueTextDocument.create(uri, 'vue', 0, content);
        }, (uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            return vscodeUri.URI.file(this.projectService.toPath(fileName)).toString();
        });
    }
    get serviceHost() {
        return this._serverHost;
    }
    get projectService() {
        return this._projectService;
    }
    log(message) {
        if (this.projectService) {
            this.projectService.logger.info(`Vue.js:: ${message}`);
        }
    }
    error(message) {
        if (this.projectService) {
            this.projectService.logger.msg(`Vue.js:: ${message} ${message.stack}`, this.typescript.server.Msg.Err);
        }
    }
    getWatchers(fileName) {
        const containingFile = vueVirtualTextdocument.isVirtualFile(fileName) ? vueVirtualTextdocument.getContainingFile(fileName) : fileName;
        if (!this.watchers.has(containingFile)) {
            this.watchers.set(containingFile, new Map());
        }
        return this.watchers.get(containingFile);
    }
    watchVirtualFile(fileName, callback) {
        const watchers = this.getWatchers(fileName);
        if (!watchers.has(fileName)) {
            watchers.set(fileName, new Set());
        }
        watchers.get(fileName).add(callback);
    }
    triggerVirtualFileWatchers(fileName, baseEvent) {
        var _a;
        const watchers = this.getWatchers(fileName);
        if (vueVirtualTextdocument.isVirtualFile(fileName)) {
            (_a = watchers.get(fileName)) === null || _a === void 0 ? void 0 : _a.forEach((callback) => {
                callback(fileName, baseEvent);
            });
        }
        else {
            watchers.forEach((watchers) => {
                watchers.forEach((callback) => {
                    callback(fileName, baseEvent);
                });
            });
        }
    }
    stopVirtualFileWatcher(fileName, callback) {
        const watchers = this.getWatchers(fileName).get(fileName);
        if (watchers) {
            watchers.delete(callback);
        }
    }
    tryCreateScriptInfo(fileName) {
        try {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                this.store.get(vueVirtualTextdocument.getContainingFile(fileName));
                const scriptInfo = this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(vueVirtualTextdocument.getContainingFile(fileName)), false);
                if (scriptInfo) {
                    patchScriptInfo(this, scriptInfo);
                }
            }
        }
        catch (_a) { }
    }
    load(info) {
        this.log(`Loading Vue plugin: ${info.project.getProjectName()}`);
        this._serverHost = info.serverHost;
        this._projectService = info.project.projectService;
        patchProjectService(this);
        patchServiceHost(this);
        patchLanguageServiceHost(this, info.languageServiceHost);
    }
}
function patchProjectService(context) {
    patchExtraFileExtensions(context);
}
function patchExtraFileExtensions(context) {
    var _a;
    const extraFileExtensions = [
        { extension: 'vue', isMixedContent: false, scriptKind: context.typescript.ScriptKind.Deferred },
    ];
    tryPatchMethod(context.projectService, 'setHostConfiguration', (setHostConfiguration) => {
        context.log(`[patch] Add support for vue extension. (ProjectService)`);
        return (args) => {
            const current = context.projectService.hostConfiguration
                .extraFileExtensions;
            if (args.extraFileExtensions) {
                args.extraFileExtensions.push(...extraFileExtensions);
                context.log(`extraFileExtensions: ${JSON.stringify(args.extraFileExtensions)}`);
            }
            else if (!current || !current.some((ext) => ext.extension === 'vue')) {
                args.extraFileExtensions = [...extraFileExtensions];
            }
            return setHostConfiguration(args);
        };
    });
    if ((_a = context.projectService.hostConfiguration.extraFileExtensions) === null || _a === void 0 ? void 0 : _a.some((ext) => ext.extension === 'vue')) {
        return;
    }
    // Enable .vue after enhancing the language server.
    setTimeout(() => context.projectService.setHostConfiguration({ extraFileExtensions: [] }), 0);
}
function patchLanguageServiceHost(context, languageServiceHost) {
    patchGetScriptFileNames(context, languageServiceHost);
    patchModuleResolution(context, languageServiceHost);
    tryPatchMethod(languageServiceHost, 'getScriptVersion', (getScriptVersion) => (fileName) => {
        if (vueVirtualTextdocument.isVirtualFile(fileName))
            return getScriptVersion(vueVirtualTextdocument.getContainingFile(fileName));
        return getScriptVersion(fileName);
    });
    tryPatchMethod(languageServiceHost, 'getScriptSnapshot', (getScriptSnapshot) => (fileName) => {
        if (vueVirtualTextdocument.isVueFile(fileName)) {
            const document = context.store.get(fileName);
            if (document) {
                return context.typescript.ScriptSnapshot.fromString(document.getDocument('_module').getText());
            }
        }
        return getScriptSnapshot(fileName);
    });
}
function patchGetScriptFileNames(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'getScriptFileNames', (getScriptFileNames) => {
        context.log(`[patch] Override getScriptFileNames to expand .vue files to virtual files. (LanguageServerHost)`);
        const previousVueFiles = new Set();
        return () => {
            const fileNames = new Set();
            [...getScriptFileNames(), ...previousVueFiles]
                .map((fileName) => (vueVirtualTextdocument.isVirtualFile(fileName) ? vueVirtualTextdocument.getContainingFile(fileName) : fileName))
                .forEach((fileName) => {
                if (vueVirtualTextdocument.isVueFile(fileName)) {
                    const document = context.store.get(fileName);
                    if (document) {
                        previousVueFiles.add(fileName);
                        fileNames.add(document.getDocumentFileName(vueVirtualTextdocument.MODULE_SELECTOR));
                    }
                    else {
                        previousVueFiles.delete(fileName);
                    }
                }
                else {
                    fileNames.add(fileName);
                }
            });
            return Array.from(fileNames);
        };
    });
}
function patchFileExists(context) {
    tryPatchMethod(context.serviceHost, 'fileExists', (fileExists = context.typescript.sys.fileExists) => {
        context.log(`[patch] Override fileExists to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                const document = context.store.get(vueVirtualTextdocument.getContainingFile(fileName));
                const { selector } = vueVirtualTextdocument.parseVirtualFileName(fileName);
                return fileName === (document === null || document === void 0 ? void 0 : document.getDocumentFileName(selector));
            }
            return fileExists(fileName);
        };
    });
}
function patchReadFile(context) {
    tryPatchMethod(context.serviceHost, 'readFile', (readFile) => {
        context.log(`[patch] Override readFile to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            var _a;
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                context.log(`host.readFile("${fileName}")`);
                const document = (_a = context.store.get(vueVirtualTextdocument.getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName);
                if (document)
                    return document.getText();
                return;
            }
            return readFile ? readFile(fileName) : context.typescript.sys.readFile(fileName);
        };
    });
}
function patchModuleResolution(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'resolveModuleNames', (resolveModuleNames) => {
        context.log(`[patch] Override resolveModuleNames to resolve imports from .vue files. (LanguageServerHost)`);
        return (moduleNames, containingFile, reusedNames, redirectedReferences, options) => {
            if (vueVirtualTextdocument.isVueFile(containingFile))
                throw new Error('A .vue file should not be part of TS program.');
            const newModuleNames = moduleNames.map((moduleName) => vueVirtualTextdocument.isVueFile(moduleName)
                ? moduleName + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + vueVirtualTextdocument.MODULE_SELECTOR
                : moduleName.endsWith('.vue?internal')
                    ? moduleName.replace(/\?internal$/, '') + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + vueVirtualTextdocument.INTERNAL_MODULE_SELECTOR
                    : moduleName);
            // TODO: Support paths mapped to .vue files, if needed.
            const result = resolveModuleNames
                ? resolveModuleNames(newModuleNames, containingFile, reusedNames, redirectedReferences, options)
                : [];
            result.map((resolved) => {
                if (resolved && vueVirtualTextdocument.isVirtualFile(resolved.resolvedFileName)) {
                    context.tryCreateScriptInfo(resolved.resolvedFileName); // Load vue file now to avoid filename case insensitivity issues.
                }
            });
            return result;
        };
    });
}
function patchServiceHost(context) {
    patchWatchFile(context);
    patchFileExists(context);
    patchReadFile(context);
}
function patchWatchFile(context) {
    tryPatchMethod(context.serviceHost, 'watchFile', (watchFile) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return (fileName, callback, pollingInterval, options) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                context.watchVirtualFile(fileName, callback);
                return {
                    close() {
                        context.stopVirtualFileWatcher(fileName, callback);
                    },
                };
            }
            return watchFile(fileName, callback, pollingInterval, options);
        };
    });
}
function patchScriptInfo(context, scriptInfo) {
    if (!scriptInfo)
        throw new Error('ScriptInfo is required.');
    tryPatchMethod(scriptInfo, 'editContent', (editContent) => {
        context.log(`[patch] Override editContent() of "${scriptInfo.fileName}" to sync virtual files. (ScriptInfo)`);
        return (start, end, newText) => {
            const document = context.store.get(scriptInfo.fileName);
            if (!document)
                throw new Error('VueTextDocument should exist for every ScriptInfo.');
            const range = { start: document.positionAt(start), end: document.positionAt(end) };
            editContent(start, end, newText);
            context.log(`UPDATE ${scriptInfo.getLatestVersion()} ==> ${scriptInfo.fileName} ${start}:${end} ${newText}`);
            vueVirtualTextdocument.VueTextDocument.update(document, [{ range, text: newText }], getLastNumberFromVersion(scriptInfo.getLatestVersion()));
            context.triggerVirtualFileWatchers(scriptInfo.fileName, context.typescript.FileWatcherEventKind.Changed);
        };
    });
}

function createServerHelper(context) {
    function getDocument(fileName) {
        var _a;
        return vueVirtualTextdocument.isVirtualFile(fileName)
            ? (_a = context.store.get(vueVirtualTextdocument.getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName) : context.store.get(fileName);
    }
    function getVueDocument(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName) ? context.store.get(fileName) : null;
    }
    function getDocumentAt(fileName, position) {
        const document = context.store.get(fileName); // The file should exist.
        const block = document.blockAt(position);
        if (!block)
            return;
        if (block.type === 'template') {
            return document.getDocument(vueVirtualTextdocument.RENDER_SELECTOR);
        }
        return document.getDocument(document.getBlockSelector(block));
    }
    function isRenderFunctionDocument(document) {
        var _a;
        return ((_a = document) === null || _a === void 0 ? void 0 : _a.selector.type) === vueVirtualTextdocument.RENDER_SELECTOR;
    }
    function isRenderFunctionFileName(fileName) {
        var _a;
        return vueVirtualTextdocument.isVirtualFile(fileName) && ((_a = vueVirtualTextdocument.parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === vueVirtualTextdocument.RENDER_SELECTOR;
    }
    return { getDocument, getDocumentAt, getVueDocument, isRenderFunctionDocument, isRenderFunctionFileName };
}

const noop = {
    getRenameInfo(fileName, position, options) {
        return {
            canRename: false,
            localizedErrorMessage: 'Not implemented.',
        };
    },
    findRenameLocations(fileName, position, findInStrings, findInComments) {
        return undefined;
    },
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return [];
    },
    getBraceMatchingAtPosition(fileName, position) {
        return [];
    },
    getBreakpointStatementAtPosition(fileName, position) {
        return undefined;
    },
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return [];
    },
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return {
            changes: [],
        };
    },
    getCompilerOptionsDiagnostics() {
        return [];
    },
    getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
        return undefined;
    },
    getCompletionEntrySymbol(fileName, position, name, source) {
        return undefined;
    },
    getCompletionsAtPosition(fileName, position, options) {
        return undefined;
    },
    getDefinitionAndBoundSpan(fileName, position) {
        return undefined;
    },
    getDefinitionAtPosition(fileName, position) {
        return undefined;
    },
    getDocCommentTemplateAtPosition(fileName, position) {
        return undefined;
    },
    getDocumentHighlights(fileName, position, filesToSearch) {
        return undefined;
    },
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        return [];
    },
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return undefined;
    },
    getEmitOutput(fileName) {
        return { outputFiles: [], emitSkipped: true };
    },
    getEncodedSemanticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getEncodedSyntacticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return [];
    },
    getFormattingEditsForDocument(fileName, options) {
        return [];
    },
    getFormattingEditsForRange(fileName, start, end, options) {
        return [];
    },
    getImplementationAtPosition(fileName, position) {
        return undefined;
    },
    getIndentationAtPosition(fileName, position, options) {
        return 0;
    },
    getJsxClosingTagAtPosition(fileName, position) {
        return undefined;
    },
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return undefined;
    },
    getNavigateToItems(searchValue) {
        return [];
    },
    getNavigationBarItems(fileName) {
        return [];
    },
    getNavigationTree(fileName) {
        return {
            text: '',
            kind: 'module',
            kindModifiers: 'component',
            spans: [],
            nameSpan: undefined,
        };
    },
    getOccurrencesAtPosition(fileName, position) {
        return undefined;
    },
    getOutliningSpans(fileName) {
        return [];
    },
    getQuickInfoAtPosition(fileName, position) {
        return undefined;
    },
    getReferencesAtPosition(fileName, position) {
        return undefined;
    },
    getSemanticClassifications(fileName, span) {
        return [];
    },
    getSemanticDiagnostics(fileName) {
        return [];
    },
    getSignatureHelpItems(fileName, position, options) {
        return undefined;
    },
    getSmartSelectionRange(fileName, position) {
        return {
            textSpan: { start: position, length: 0 },
        };
    },
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return undefined;
    },
    getSuggestionDiagnostics(fileName) {
        return [];
    },
    getSyntacticClassifications(fileName, span) {
        return [];
    },
    getSyntacticDiagnostics(fileName) {
        return [];
    },
    getTodoComments(fileName, descriptors) {
        return [];
    },
    getTypeDefinitionAtPosition(fileName, position) {
        return [];
    },
    findReferences(fileName, position) {
        return undefined;
    },
    prepareCallHierarchy(fileName, position) {
        return undefined;
    },
    provideCallHierarchyIncomingCalls(fileName, position) {
        return [];
    },
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return [];
    },
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return false;
    },
    organizeImports(scope, formatOptions, preferences) {
        return [];
    },
    getProgram() {
        throw new Error('Not Implemented');
    },
    cleanupSemanticCache() {
        throw new Error('Not Implemented');
    },
    dispose() {
        throw new Error('Not Implemented');
    },
    applyCodeActionCommand() {
        return Promise.reject('Not Implemented');
    },
};

function createTemplateLanguageServer({ helpers: h, service, context, }) {
    function getRenderDoc(fileName) {
        return h.getDocument(fileName);
    }
    function getTemplateDoc(fileName) {
        return getRenderDoc(fileName).container.getDocument('template');
    }
    function getTextSpan(document, span) {
        if (h.isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result)
                return { start: result.offset, length: result.length };
        }
        return span;
    }
    return {
        ...noop,
        getRenameInfo(fileName, position, options) {
            var _a;
            const document = getRenderDoc(fileName);
            const template = getTemplateDoc(fileName);
            const offset = (_a = document.getGeneratedOffsetAt(position)) === null || _a === void 0 ? void 0 : _a.offset;
            if (vueVirtualTextdocument.isNumber(offset)) {
                const result = service.getRenameInfo(fileName, offset, options);
                if (result.canRename) {
                    result.triggerSpan = getTextSpan(document, result.triggerSpan);
                    const prefix = template.getText(result.triggerSpan.start - 1, 1);
                    if (result.displayName === '$event') {
                        return {
                            canRename: false,
                            localizedErrorMessage: '$event is builtin variable, it cannot be renamed.',
                        };
                    }
                    else if (prefix === ':' || prefix === '@') {
                        return {
                            canRename: false,
                            localizedErrorMessage: (prefix === ':' ? 'Prop/attribute' : 'Event name') + ' renaming is not supported.',
                        };
                    }
                }
                return result;
            }
            return {
                canRename: false,
                localizedErrorMessage: 'You cannot rename this element.',
            };
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a;
            const document = getRenderDoc(fileName);
            const offset = (_a = document.getGeneratedOffsetAt(position)) === null || _a === void 0 ? void 0 : _a.offset;
            if (vueVirtualTextdocument.isNumber(offset)) {
                return service.findRenameLocations(fileName, offset, findInStrings, findInComments);
            }
            return undefined;
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            const document = getRenderDoc(fileName);
            // TODO: ...
            if (vueVirtualTextdocument.isNumber(positionOrRange)) {
                const offset = document.getGeneratedOffsetAt(positionOrRange);
                if (!offset)
                    return [];
                positionOrRange = offset.offset;
            }
            else {
                const offset = document.getGeneratedOffsetAt(positionOrRange.pos);
                if (!offset)
                    return [];
                const diff = positionOrRange.end - positionOrRange.pos;
                positionOrRange.pos = offset.offset;
                positionOrRange.end = offset.offset + Math.min(offset.length, Math.max(1, diff));
            }
            return service.getApplicableRefactors(fileName, positionOrRange, preferences);
        },
    };
}

function createVueLanguageServer(options) {
    const template = createTemplateLanguageServer(options);
    const { helpers: h, service: script, context } = options;
    function choose(document) {
        return h.isRenderFunctionDocument(document) ? template : script;
    }
    return {
        ...noop,
        getRenameInfo(fileName, position, options) {
            const document = h.getDocumentAt(fileName, position);
            if (!document) {
                return {
                    canRename: false,
                    localizedErrorMessage: '',
                };
            }
            return choose(document).getRenameInfo(document.fsPath, position, options);
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            const document = h.getDocumentAt(fileName, position);
            if (!document)
                return;
            return choose(document).findRenameLocations(document.fsPath, position, findInStrings, findInComments);
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            const document = h.getDocumentAt(fileName, vueVirtualTextdocument.isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos);
            if (document) {
                return choose(document).getApplicableRefactors(document.fsPath, positionOrRange, preferences);
            }
            return [];
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            const document = h.getVueDocument(oldFilePath);
            if (document) {
                const doc = document.getDocument('_module');
                return script.getEditsForFileRename(doc.fsPath, doc.fsPath.replace(oldFilePath, newFilePath), formatOptions, preferences);
            }
            return [];
        },
    };
}

const VUE_LANGUAGE_SERVER = Symbol('Vue Language Server');
class RoutingLanguageServer {
    constructor(context) {
        this.context = context;
    }
    decorate(languageService) {
        if (VUE_LANGUAGE_SERVER in languageService) {
            return languageService;
        }
        const proxy = createLanguageServiceRouter({
            context: this.context,
            service: languageService,
            helpers: createServerHelper(this.context),
        });
        // @ts-ignore
        proxy[VUE_LANGUAGE_SERVER] = true;
        return proxy;
    }
}
function createLanguageServiceRouter(options) {
    const vue = createVueLanguageServer(options);
    const ts = options.service;
    function choose(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName) ? vue : ts;
    }
    const methods = [
        'getRenameInfo',
        'findRenameLocations',
        'getApplicableRefactors',
        'getBraceMatchingAtPosition',
        'getBreakpointStatementAtPosition',
        'getCodeFixesAtPosition',
        'getCompletionEntryDetails',
        'getCompletionEntrySymbol',
        'getCompletionsAtPosition',
        'getDefinitionAndBoundSpan',
        'getDefinitionAtPosition',
        'getDocCommentTemplateAtPosition',
        'getDocumentHighlights',
        'getEditsForFileRename',
        'getEditsForRefactor',
        'getEmitOutput',
        'getEncodedSemanticClassifications',
        'getEncodedSyntacticClassifications',
        'getFormattingEditsAfterKeystroke',
        'getFormattingEditsForDocument',
        'getFormattingEditsForRange',
        'getImplementationAtPosition',
        'getIndentationAtPosition',
        'getJsxClosingTagAtPosition',
        'getNameOrDottedNameSpan',
        'getNavigateToItems',
        'getNavigationBarItems',
        'getNavigationTree',
        'getOccurrencesAtPosition',
        'getOutliningSpans',
        'getQuickInfoAtPosition',
        'getReferencesAtPosition',
        'getSemanticClassifications',
        'getSemanticClassifications',
        'getSignatureHelpItems',
        'getSmartSelectionRange',
        'getSpanOfEnclosingComment',
        'getSuggestionDiagnostics',
        'getSyntacticDiagnostics',
        'getTodoComments',
        'getTypeDefinitionAtPosition',
        'findReferences',
        'prepareCallHierarchy',
        'provideCallHierarchyIncomingCalls',
        'provideCallHierarchyOutgoingCalls',
        'isValidBraceCompletionAtPosition',
    ];
    function getTextSpan(document, span) {
        if (options.helpers.isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result)
                return { start: result.offset, length: result.length };
            return null;
        }
        return span;
    }
    const proxy = {
        organizeImports(scope, formatOptions, preferences) {
            return choose(scope.fileName).organizeImports(scope, formatOptions, preferences);
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a;
            return (_a = choose(fileName)
                .findRenameLocations(fileName, position, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.map((item) => {
                options.context.log('xxx.findRenameLocations ' + JSON.stringify(item));
                if (vueVirtualTextdocument.isVirtualFile(item.fileName)) {
                    item.originalContextSpan = item.contextSpan;
                    item.originalTextSpan = item.textSpan;
                    item.originalFileName = item.fileName;
                    const virtual = options.helpers.getDocument(item.fileName);
                    item.fileName = virtual.container.fsPath;
                    const textSpan = getTextSpan(virtual, item.textSpan);
                    if (!textSpan)
                        return;
                    item.textSpan = textSpan;
                    if (item.contextSpan) {
                        const contextSpan = getTextSpan(virtual, item.contextSpan);
                        if (!contextSpan)
                            return;
                        item.contextSpan = contextSpan;
                    }
                }
                return item;
            }).filter(isNotNull);
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            const suffix = '.vue' + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + '_module';
            return choose(oldFilePath)
                .getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences)
                .map((edit) => {
                if (vueVirtualTextdocument.isVirtualFile(edit.fileName)) {
                    const selector = vueVirtualTextdocument.parseVirtualFileName(edit.fileName);
                    if ((selector === null || selector === void 0 ? void 0 : selector.selector.type) !== 'script')
                        return;
                    edit.fileName = vueVirtualTextdocument.getContainingFile(edit.fileName);
                }
                edit.textChanges = edit.textChanges
                    .map((change) => {
                    const index = change.newText.indexOf(suffix);
                    if (index >= 0) {
                        change.newText = change.newText.substr(0, index + 4) + change.newText.substr(index + suffix.length);
                    }
                    return change;
                })
                    .filter(isNotNull);
                if (!edit.textChanges.length)
                    return;
                return edit;
            })
                .filter(isNotNull);
        },
    };
    methods.forEach((name) => {
        if (!(name in proxy)) {
            // @ts-ignore
            proxy[name] = function (fileName) {
                const service = choose(fileName);
                // @ts-ignore
                return service[name].apply(service, arguments);
            };
        }
    });
    return proxy;
}
function isNotNull(value) {
    return value != null;
}

let context;
let server;
function init({ typescript }) {
    if (!context) {
        context = new PluginContext(typescript);
        server = new RoutingLanguageServer(context);
    }
    return {
        create(info) {
            const version = findVueVersionIn(info.project.getCurrentDirectory());
            if (!isVue3(version)) {
                return info.languageService;
            }
            else {
                context.load(info);
                return server.decorate(info.languageService);
            }
        },
    };
}
function isVue3(version) {
    if (!version)
        return true; // assume Vue 3 by default.
    return version.startsWith('3.');
}
function findVueVersionIn(directory) {
    var _a, _b;
    const packageFile = findNearestPackageFile(directory);
    if (packageFile) {
        const vuePackageFile = Path.resolve(Path.dirname(packageFile), 'node_modules', 'vue', 'package.json');
        if (FS.existsSync(vuePackageFile)) {
            return require(vuePackageFile).version;
        }
        // dependencies not installed.
        const pkg = require(packageFile);
        const version = ((_a = pkg.dependencies) === null || _a === void 0 ? void 0 : _a.vue) || ((_b = pkg.devDependencies) === null || _b === void 0 ? void 0 : _b.vue);
        if (!version)
            return;
        const justNumbers = version.replace(/[^0-9.]/, '');
        if (justNumbers.startsWith('3.'))
            return justNumbers;
        if (version === 'next')
            return '3.0.0';
        return '2.x';
    }
}
function findNearestPackageFile(directory) {
    if (!directory)
        return;
    const packageFile = Path.resolve(directory, 'package.json');
    if (FS.existsSync(packageFile))
        return packageFile;
    if (directory === '/')
        return;
    return findNearestPackageFile(Path.dirname(directory));
}

module.exports = init;
//# sourceMappingURL=typescript-plugin-vue.js.map
